/***************************************************************************
 *  colitas.cc : This file is part of 'colitas'
 *
 *  (c) 2004 by Lukasz Tomicki <tomicki@o2.pl>
 *
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "colitas.h"
#include "new_on.h"
#include "MemoryManager.h"

using namespace std;

CScanner::CScanner() : ports_first(false), scan_speed(150), wait_seconds(3)
{
	output_func = 0;
}

void CScanner::set_output(void (*p)(char const*))
{
	output_func = p;
}

void CScanner::get_options(int argc, char **argv)
{
	char o;
	
	while ((o = getopt(argc, argv, "t:ps:")) >= 0) {
		switch (o) {
		case 's':
			wait_seconds = atoi(optarg);
		break;
		
		case 'p':
			ports_first = true;
		break;
		
		case 't':
			scan_speed = atoi(optarg);
		break;
		
		default:
			break;
		}
	}
}

u32 CScanner::get_ports()
{
	return ( ports.size() );
}

u32 CScanner::get_hosts()
{
	return ( (input_data.e4 - input_data.s4 + 1) * 
	(input_data.e3 - input_data.s3 + 1) * 
	(input_data.e2 - input_data.s2 + 1) * 
	(input_data.e1 - input_data.s1 + 1) );
}

void CScanner::populate_ip_addr(const char *p)
{
	bool found(false);
	for (int i(0); p[i]; ++i) {
		if (p[i] == ':') {
			found = true;
			break;
		}
	}
	input_data.e1 = 0;
	if (found) {
		sscanf(p, "%u.%u.%u.%u:%u.%u.%u.%u", &input_data.s1,  &input_data.s2, 
		 &input_data.s3,  &input_data.s4, &input_data.e1,  &input_data.e2, 
		 &input_data.e3,  &input_data.e4);
	} else {
		sscanf(p, "%u.%u.%u.%u-%u.%u.%u.%u", &input_data.s1,  &input_data.s2, 
		 &input_data.s3,  &input_data.s4, &input_data.e1,  &input_data.e2, 
		 &input_data.e3,  &input_data.e4);
	}
	if (!input_data.e1) {
		input_data.e1 = input_data.s1;
		input_data.e2 = input_data.s2;
		input_data.e3 = input_data.s3;
		input_data.e4 = input_data.s4;
	}
}

void CScanner::populate_port_list(const char *p)
{
	ports.clear();
	//puts(p);
	char *buffer = new char[strlen(p) + 1];
	int b(0);
	u32 t, s(0), e;
	for (int i(0); p[i]; ++i) {
		if (p[i] == ',') {
			++b;
			buffer[b] = 0;
			if (s) {
				e = atoi(buffer);
				if (e >= s) {
					for (; s <= e; ++s)
						ports.push_back(s);
				}
				s = 0;
			} else {
				t = atoi(buffer);
				if (t > max_port)
					t = max_port;
				ports.push_back(t);
			}
			b = 0;
		} else if (p[i] == '-' || p[i] == ':') {
			++b;
			buffer[b] = 0;
			s = atoi(buffer);
			if (!s)
				s = 1;
			b = 0;
		} else if (p[i] != ' ') {
			buffer[b] = p[i];
			++b;
		}
	}
	++b;
	buffer[b] = 0;
	if (s) {
		e = atoi(buffer);
		if (e > max_port)
			e = max_port;
		
		if (e >= s) {
			for (; s <= e; ++s)
				ports.push_back(s);
		}
		s = 0;
	} else {
		t = atoi(buffer);
		ports.push_back(t);
	}
	delete [] buffer;
}

unsigned long CScanner::get_next_port_host()
{
	if (!out_data.port)
		out_data.port = get_next_port();
	if (out_data.s4 == input_data.e4 || out_data.s4 == 255) {
		out_data.s4 = 1;
		if (out_data.s3 == input_data.e3 || out_data.s3 == 255) {
			out_data.s3 = 0;
			if (out_data.s2 == input_data.e2 || out_data.s2 == 255) {
				out_data.s2 = 0;
				if (out_data.s1 == input_data.e1) {
					out_data.port = get_next_port();
					if (!out_data.port)
						return 0;
					else {
						out_data.s1 = input_data.s1;
						out_data.s2 = input_data.s2;
						out_data.s3 = input_data.s3;
						out_data.s4 = input_data.s4;
					}
				} else
					++out_data.s1;
			} else
				++out_data.s2;
		} else 	
			++out_data.s3;
	} else
		++out_data.s4;
	
	char buffer[32];
	if (!out_data.s4)
		++out_data.s4;
	
	sprintf(buffer, "%d.%d.%d.%d", out_data.s1, out_data.s2, out_data.s3, out_data.s4);
	//printf("%s %d\n", buffer, out_data.port);
	return ( inet_addr(buffer) );
}

unsigned long CScanner::get_next_host_port()
{
	out_data.port = get_next_port();
	if (!out_data.port) {
		out_data.port = get_next_port();
		if (out_data.s4 == input_data.e4 || out_data.s4 == 255) {
			out_data.s4 = 1;
			if (out_data.s3 == input_data.e3 || out_data.s3 == 255) {
				out_data.s3 = 0;
				if (out_data.s2 == input_data.e2 || out_data.s2 == 255) {
					out_data.s2 = 0;
					if (out_data.s1 == input_data.e1)
						return 0;
					else
						++out_data.s1;
				} else
					++out_data.s2;
			} else 	
				++out_data.s3;
		}
		++out_data.s4;
	}
	
	char buffer[32];
	if (!out_data.s4)
		++out_data.s4;
	
	sprintf(buffer, "%d.%d.%d.%d", out_data.s1, out_data.s2, out_data.s3, out_data.s4);
	//printf("%s %d\n", buffer, out_data.port);
	return ( inet_addr(buffer) );
}

unsigned long CScanner::get_own_ip()
{
	unsigned long s_addr(0);
	int sock;
	struct ifreq ifreq;
    struct sockaddr_in *saptr = 0;
    struct if_nameindex *listsave = 0;
      
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
        return 0;
	}
	
	listsave = if_nameindex();
	
    for(struct if_nameindex *iflist = listsave; iflist->if_name != 0; ++iflist) {
		strncpy(ifreq.ifr_name, iflist->if_name, IF_NAMESIZE);
		if(ioctl(sock, SIOCGIFADDR, &ifreq) != 0)
			continue;
		
		saptr = (struct sockaddr_in *)&ifreq.ifr_addr;
		s_addr = saptr->sin_addr.s_addr;
    }
     
	if_freenameindex(listsave);
	return ( s_addr );
}

u16 CScanner::get_next_port()
{
	static vector<u16>::iterator i = ports.begin();
	if (i == ports.end()) {
		i = ports.begin();
		return 0;
	}
	u16 p = *i;
	++i;
	return p;
}
	
bool CScanner::start_scan()
{
	if (!check_input_data(input_data))
		return false;
	
	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_TCP);
  	if (sockfd < 0) {
		perror("cannot create socket");
   		return false;
 	}
	
  	int on(1);
 	if (setsockopt(sockfd, IPPROTO_IP, IP_HDRINCL, (char*)&on, 
		sizeof(on)) == -1) {
			perror("cannot setservaddr");
			return false;
	}

	out_data = input_data;
	if (!ports_first && input_data.s4)
		--out_data.s4;
	srand(getpid());
	ack_code = htonl(rand());
	seq_ack = ack_code + htonl(1);
	unsigned long d_addr;
	unsigned long s_addr = get_own_ip();
	while (d_addr = get_next_addr()) {
		if (!syn_send(s_addr, d_addr, rand(),out_data.port)) {
			printf("error while sending SYN packets\n");
			printf("socket timed out\n");
		}
		syn_recv(scan_speed);
	}
	syn_recv(0, wait_seconds);
	
	close(sockfd);
	return true;
}

inline unsigned long CScanner::get_next_addr()
{
	if (!ports_first)
		return ( get_next_port_host() );
	
	return ( get_next_host_port() );
}

bool CScanner::check_input_data(scan_data input)
{
	if (input.s1 < input.e1)
		return true;
	if (input.s1 > input.e1)
		return false;
	if (input.s2 < input.e2)
		return true;
	if (input.s2 > input.e2)
		return false;
	if (input.s3 < input.e3)
		return true;
	if (input.s3 > input.e3)
		return false;
	if (input.s4 < input.e4)
		return true;
	if (input.s4 > input.e4)
		return false;
	return true;
}

unsigned short CScanner::in_cksum(unsigned short *ptr, int nbytes)
{
  register long	sum;
  u_short oddbyte;
  register u_short answer;

  sum = 0;
  while(nbytes > 1) {
    sum += *ptr++;
    nbytes -= 2;
  }

  if(nbytes == 1) {
    oddbyte = 0;
    *((u_char *) &oddbyte) = *((u_char *)ptr);
    sum += oddbyte;
  }

  sum  = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  answer = ~sum;

  return ( answer );
}

bool CScanner::syn_send(unsigned int saddr, unsigned int daddr, unsigned short sport,
		unsigned short dport, u32 timeout)
{
	static char packet[sizeof(iphdr) + sizeof(tcphdr)];
	memset(packet, 0, sizeof(iphdr) + sizeof(tcphdr));
	
	sockaddr_in servaddr;
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(dport);
	servaddr.sin_addr.s_addr = daddr;
	memset(&servaddr.sin_zero, 0, sizeof(servaddr.sin_zero));
	
	pseudohdr *pseudo = (pseudohdr *)(packet + sizeof(iphdr) - sizeof(pseudohdr));
	pseudo->saddr = saddr;
	pseudo->daddr = daddr;
	pseudo->protocol = IPPROTO_TCP;
	pseudo->length = htons(sizeof(tcphdr));
	
	tcphdr *tcp = (tcphdr *)(packet + sizeof(iphdr));
	tcp->source = htons(sport);
	tcp->dest= htons(dport);
	tcp->seq = ack_code;
	tcp->ack_seq = rand();
	tcp->doff = 5;
	tcp->syn = 1;
	tcp->window = htons (512);
	tcp->check = in_cksum((unsigned short *)pseudo, sizeof(tcphdr) + sizeof(pseudohdr));

	iphdr *ip = (iphdr *)packet;
 	ip->version = 4;
 	ip->ihl = 5;
 	ip->tos = 0x10;
 	ip->tot_len = htons(sizeof(iphdr) + sizeof(tcphdr));
 	ip->id = rand();
 	ip->ttl = 255;
 	ip->protocol = IPPROTO_TCP;
 	ip->saddr = saddr;
 	ip->daddr = daddr;
 	ip->check = in_cksum((unsigned short *)ip, sizeof(iphdr));
	
	timeval *ptv = 0;
	timeval tv;
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	
	ptv = &tv;

	fd_set writefds;

	FD_ZERO(&writefds);
	FD_SET(sockfd, &writefds);
	select(sockfd + 1, 0, &writefds, 0, ptv);
	if (!FD_ISSET(sockfd, &writefds))
		return false;
	
 	if ((sendto(sockfd, packet, sizeof(iphdr) + sizeof(tcphdr), 0, 
		(sockaddr*)&servaddr, sizeof(sockaddr_in))) == -1) {
		printf(" host: %s\n", inet_ntoa(servaddr.sin_addr));
		perror("sendto");
	}
   	
	return true;
}

void CScanner::syn_recv(u32 timeout, u32 sec)
{
	static char buffer[sizeof(iphdr) + sizeof(tcphdr)];
	fd_set readfds;
	struct timeval tv;
	tv.tv_sec = sec;
    tv.tv_usec = timeout;
	bool work_on;
	FD_ZERO(&readfds);
	FD_SET(sockfd, &readfds);
   	select(sockfd + 1, &readfds, 0, 0, &tv);
	if (FD_ISSET(sockfd, &readfds))
		work_on = true;
	else
		work_on = false;
		
	tcphdr *tcp = (tcphdr*)(buffer + sizeof(iphdr));
	while (work_on) {
		if (sizeof(buffer) != recv(sockfd, buffer, sizeof(buffer), 0))
			return;
		
		if (tcp->ack_seq == seq_ack) {
			if (output_func)
				(*output_func)(buffer);
		
			tv.tv_sec = sec;
    		tv.tv_usec = timeout;
		}
		FD_ZERO(&readfds);
		FD_SET(sockfd, &readfds);
   	 	select(sockfd + 1, &readfds, 0, 0, &tv);
		if (FD_ISSET(sockfd, &readfds))
			work_on = true;
		else
			work_on = false;
	}
}
